;
; System Stabilizer for Fargo II v0.7
;
; By Bryan Turner
; 
;
; Fargo II calls ER_catch and saves the state of the machine before calling a program
;    this is very usefull since at any time we can exit from the program by calling ER_throw
;    all we have to do is execute one simple instruction: $A001
;
; There are some situations where exiting from the programs could be bad, but we'll ignore
;    those for now.
;
; Features:	Turn off grayscale			- v0.5
;		Re-adjust the LCD base addr		- v0.5
;		Turn off Hardware Grayscale		- v0.5
;		Set memory protection			- v0.5
;		Adjust Status Register to normal	- v0.6 + even smaller!
;		Does not rely on valid stack pointer	- v0.7 + beginning optimization...
;		Copy to Memory Protected Area		- v0.7 + About as small as I can make it
; Future?
;		Unload used memory?
;		Adjust all interrupts back to normal?

	include "flib.h"
	include "tios.h"
	xdef	_main
	xdef	_comment


NUMTRAPS	equ	16


_comment:
	dc.b	"System Stabilizer Installer",0
	ds.w	0

_main:
	jsr	flib::clr_scr
	
	move.w	#1,-(a7)
	jsr	tios::FontSetSys	; 8pt font
	addq	#2,a7

	move.w	#$700,d0		; disable interrupts, and put
	trap	#1			; the old int mask in D0
	move.w	d0,d7			; save old mask
		
	move.w  ($8),d0			; Check that Raven or another debugger is not installed..
	cmp.w	#$0040,d0
	bne	debug_there		; Something's already there!
	
	bclr.b	#2,($600001)		; Memory protect OFF
	
	lea	(Mem_protect+2)(pc),a0
	move.w	($600000),(a0)		; Store normal, non-memory protected data from $600000

	lea	(auto_one+2)(pc),a0
	move.l	($64),(a0)		; Store original auto int 1 proc (for use in exception handler)
	
	lea	(LCD_addr+2)(pc),a0
	move.l	#tios::main_lcd,d0	; Load basic LCD address for this calc
	lsr.l	#3,d0
	move.w	d0,(a0)
;
; BT: Now copied to unused protected vector space at address $C0-$100
;
;	move.l	($5d36),a1
;	move.l	#the_end-theProc,d0	; Load size of program
;	sub.l	d0,a1			; Subtract size
;	move.l	a1,($5d36)		; Store new top of heap --- NOT ROM 2.1 Compatable
;	move.l	a1,d1			; Keep a copy in D1
;
	; BT: Copy to unused vector space...
	movea.w	#$C0,a1			; Load address to start at
	move.l	a1,d1			; Keep a copy in D1
	move.l	#the_end-theProc,d0	; Load size of program

	lea	theProc(pc),a0		; Load address to start of handler
	lsr.l	#2,d0			; Divide size by 4 (number of longwords)
	subq	#1,d0			; Adjust to 0..numlongs-1
copy_loop:
	move.l	(a0)+,(a1)+		; Copy exception processor to high memory
	dbra	d0,copy_loop		; *** May not copy last 2 bytes!!

	;
	; Install exception proc to catch lots of errors
	;
	lea	TRAPNUMS(pc),a0		; Load trap numbers to catch
	moveq	#NUMTRAPS-1,d2		; Load number of traps to edit
	clr.l	d0

inlup:
	move.b	(a0)+,d0		; Get a trap number
	move.l	d0,a1			; Make it an address
	move.l	d1,(a1)			; Install our exception routine
	dbra	d2,inlup		; Loop back..
	
	bset.b	#2,($600001)		; Memory protect ON
	
done:
	move.w	#4,-(a7)
	pea	good_str(pc)		; Draw good_str
	move.l	#$00100010,-(a7)
	jsr	tios::DrawStrXY
	lea	10(a7),a7

	move.w	#4,-(a7)
	pea	good_str2(pc)		; Draw good_str2
	move.l	#$0010001A,-(a7)
	jsr	tios::DrawStrXY
	lea	10(a7),a7

	; --- Delete the program!  It is no longer needed, and only taking up space!
	
	bra	pause
debug_there:
	move.w	#4,-(a7)
	pea	debugger(pc)		; Draw Debuuger str
	clr.l	-(a7)
	jsr	tios::DrawStrXY
	lea	10(a7),a7
	bra	bad_install
	
bad_install:
	move.w	#4,-(a7)
	pea	bad_str(pc)
	move.w	#10,-(a7)
	clr.w	-(a7)
	jsr	tios::DrawStrXY		; Draw bad_str
	lea	10(a7),a7

pause:
	move.w	d7,d0
	trap	#1			; Re-enable interrupts

	jsr	flib::idle_loop		; Wait for a keypress
	rts
	
	;
	; The trap numbers we replace
	;
TRAPNUMS:
	dc.b	$08,$0c,$10,$14,$18,$1c,$20,$24,$2c
	dc.b	$60,$6c,$7c,$ac,$b0,$b4,$b8,0
	EVEN
debugger:
	dc.b	"Debugger already present.",0
	EVEN
bad_str:
	dc.b	"Not Installed!",0
	EVEN
good_str:
	dc.b	"Instalation Complete, you may",0
	EVEN

good_str2:
	dc.b	"now delete this program!",0,0
	EVEN

theProc:
;
; ----------------------------------------------------------
; Everything below this line is copied to a new place in memory.
;

	;
	; This is the exception processor, simply stabilize the system, and call ER_Throw
	;
	; BT: To effect the smallest possible footprint, I've begun to optimize the code
	;     sorry it's unreadable now, but be happy - it's nearly half the original length!
	;
except:
	move.w	#$2700,sr		; Disable interrupts
	move.l	#$600000,a0
LCD_addr:
	move.l	#$08883180,$10(a0)	; Turn off HW Grayscale, and reset LCD to base buffer
Mem_protect:
	move.l	#$A40500DE,(a0)		;Turn off memory protect and reset HW grayscale settings
auto_one:
	move.l	#$10000001,($64)	; Restore correct autoint 1 (Turns off GRAY screens)	
	bset.b	#2,(a0)			; Memory protect ON
	move.w	#0,sr			; Enable interrupts, force User Mode
throw:
	tios::ER_throw	1		; Error #1; Pop system state and return to Fargo


;
; Do not add anything below this line.
; ------------------------------------------
;
padding:
	dc.w	0
the_end:
	dc.w	0

	end