//
// Simple C-Level Implementation of Fargo Programming
//
// Bryan Turner

//
// This program shows how several interesting and powerful features of the TI-OS
//    can be accessed through C without needing hard-core assembly experience.
//

// --- HEADER FILES ---
#include "TIOS.h"
#include "TI-RAM.h"
#include "Globals.h"


// --- THE HEADER.C FILE ---
//#include "Header2.c"


// --- LIBRARY GLUE.C FILES ---
//#include "TIOSGlue.c"

Globals myGlobals;	// This MUST be the only global variable or strange things will occur!


void WaitTimer( short howLong, char whichTimer )
{

	// Setup timer 5 (may throw another process off if it is already using timer 5!!)
	TI_timers[whichTimer].InitVal = 0;
	TI_timers[whichTimer].CurVal = howLong;
	
	// Wait on the timer...
	while( TI_Timer_Expired[whichTimer] == 0 )
		;
	
	// Release the timer for another process to use
	TI_timers[whichTimer].InitVal = 0xFFFFFFFF;
	TI_Timer_Expired[whichTimer] = 0;
	
}

// This is the first procedure called, like 'main' in a normal program
// You may use other procedures or lump all the code in here.
void myCode( void )
{
	int x;
	long * data = (long *)&TI_BeginLCDBuff;
	
	// Fill the screen with vertical stripes
	for ( x=0x3C0; x > 0; x-- )
		*(data++) = 0xCCCCCCCC;
	
	// Set the activity indicator to 'BUSY'
	ST_busy(BUSY);
	ST_redraw();
	
	// Print some text using the ROM function 'puttext'
	DrawStrXY( 8, 8, "--PRESS A KEY--", DIMMED );
	
	// Wait for a key..
	while ( TI_KeyBuffer.used == 0 )
		;
	
	// Get the character in the buffer, put it in a global
	myGlobals.ch = TI_KeyBuffer.buffer[0];
	TI_KeyBuffer.used = 0;
	
	// Some more text..
	DrawStrXY( 8, 18, "You typed:", PLAIN );
	DrawCharXY( myGlobals.ch, 90, 18, INVERSE, 0xFF, 0, 0xFF );
	DrawStrXY( 8, 30, "Waiting on the Timer...", PLAIN );
	
	// Wait on timer 5 for 0x100 cycles...
	WaitTimer( 0x60, 5 );
	
	// Set the activity indicator to 'IDLE'
	ST_busy(IDLE);
	ST_redraw();
	
}

//#include "Main2.c"

