//
//  Final Demo Program
//
// Bryan Turner

#include <XinuProc.h>
#include <TIOS.h>

typedef struct
	{
	short	theHandle;		// Handle of the data to copy
	short	beginOffset;	// Offset of process to run
	} LoadStruct;
	
typedef struct
	{
	short *screenBuff;		// Xinu Screen Buffer
	short *bufferMask;		// Xinu Screen Mask
	
	SEM_ID maskSem;			// Screen Mask Semaphore
	
	} ImportantStuff;

typedef struct	// Sprite 16x16 pixels embedded in 32x16 bit matrix
	{
	long *ScreenPtr;
	long *BackPtr;
	short *maskPtr;
	short curMask;		// Current shorts bieng used on each scanline
	
	long Pic[17];		// First row is mask to determine when to advance screen pointer 
	short xOff, yOff;	// Offset in each direction on next move
	short sleepTime;	// Time before next movement (1/10 seconds)
	} Sprite;

typedef struct
	{
	char	height, width;	// Window height must be > 8, width specified in multiples of 16 (1,2,3..)
	short	windowMask,x,y;
	short	*screenPtr;
	short	*maskPtr;
	} XinuWindow;


asm void BufferMask()	// High bit of each short not used!!
{
	dc.w	0xFFFF		// Border for mask... (for sprites & windows)
	ds.w	128
	dc.w	0xFFFF		// Border for sprites & windows
}

asm void ScreenBuffer()
{
	ds.w	128*15
	ds.l	2				// Guarantees an 8-byte boundary - which is required..
}


asm void CopyScreen( char *buffMask, char *screenBuff )
{
	fralloc
	
copyLup:
	movea.l	buffMask,a2		// Load Screen Mask address
	movea.l	screenBuff,a1	// Load Xinu Screen address
	movea.w	%0x4440,a0		// Load TI-Screen address
	
	move.w	%127,d0
copyAll:
	move.w	(a2)+,d2
	bne		slowCopy		// Is this line masked?
	// Fast Copy!
	
	movem.l	(a0)+,D1-D7
	movem.l	D1-D7,(a1)		// Copy entire line! (well, all but 2 bytes)
	adda.w	%28,a1
	move.w	(a0)+,(a1)+		// Copy last two bytes of line..
	
	dbra	d0,copyAll
	bra		copyDone
	
slowCopy:
	
	moveq	%14,d1			// Start at bit #15
copyLine:
	
	btst	d1,d2			// Test the bit
	beq		copyIt			// BEQ : Branch if bit is == 0
	
	addq	%2,a0			// Masked: Advance pointers
	addq	%2,a1
	//move.w	%0xFFFF,(a1)+	// --- TESTING: Fill in the masked area
	
	dbra	d1,copyLine		// Loop back
	
	dbra	d0,copyAll
	bra		copyDone
	
copyIt:
	// Not masked: copy the word
	move.w	(a0)+,(a1)+
	dbra	d1,copyLine
	
	dbra	d0,copyAll

copyDone:
	
	frfree
	rts

}

asm void XinuSprite()
{
	// BT
	// Sprite is only 16x16, but embeded in 32x16 for code efficency.
	//
	dc.l	0x00FFFF00	// Sprite mask, all 16 bits of sprite width must be represented on this line
	
	dc.l	0x00FFFF00
	dc.l	0x00CCCC00
	dc.l	0x00FFFF00
	dc.l	0x00FDF900
	
	dc.l	0x00FC7500
	dc.l	0x00FE6D00
	dc.l	0x00FE6F00
	dc.l	0x00FE1F00
	
	dc.l	0x00BF3F00
	dc.l	0x009F1F00
	dc.l	0x008F0F00
	dc.l	0x00C0C300
	
	dc.l	0x00E3FF00
	dc.l	0x00FFFF00
	dc.l	0x00CCCC00
	dc.l	0x00FFFF00
	}

asm void TISprite()
{
	// BT
	// Sprite is only 16x16, but embeded in 32x16 for code efficency.
	//
	dc.l	0x00FFFF00	// Sprite mask, all 16 bits of sprite width must be represented on this line
	
	dc.l	0x000F0000
	dc.l	0x000F0000
	dc.l	0x000F2000
	dc.l	0x000F8C00
	
	dc.l	0x000FAC00
	dc.l	0x000E2600
	dc.l	0x00FE4600
	dc.l	0x00FFDF00
	
	dc.l	0x007F7F00
	dc.l	0x003F5E00
	dc.l	0x003A1C00
	dc.l	0x00131C00
	
	dc.l	0x0003F800
	dc.l	0x0001E000
	dc.l	0x0000E000
	dc.l	0x00002000
	}

void MoveSprite( Sprite *theSprite )
{
	short idx, *walk;
	long *longwalk;
	
	// First, move sprite up or down...
	if ( theSprite->yOff != 0 )
		{
		// Test for a bounce off another object...
		if ( theSprite->yOff < 0 )
			{
			if ( ((*(short *)(((char *)theSprite->maskPtr)-2)) & (theSprite->curMask)) != 0 )
				theSprite->yOff = 1;
			}
		else if ( ((*(short *)(((char *)theSprite->maskPtr)+32)) & (theSprite->curMask)) != 0 )
			theSprite->yOff = -1;
		
		
		if ( theSprite->yOff < 0 )
			{
			(char *)theSprite->ScreenPtr -= 30;					// Adjust drawing pointer
			(char *)theSprite->BackPtr   -= 30;
			*(short *)(((char *)theSprite->maskPtr)-2)  |= theSprite->curMask;
			*(short *)(((char *)theSprite->maskPtr)+30) &= ~(theSprite->curMask);
			(char *)theSprite->maskPtr   -= 2;					// Adjust mask pointer
			}
		else
			{
			(char *)theSprite->ScreenPtr += 30;					// Adjust drawing pointer
			(char *)theSprite->BackPtr   += 30;
			(char *)theSprite->maskPtr   += 2;					// Adjust mask pointer
			*(short *)(((char *)theSprite->maskPtr)-2)  &= ~(theSprite->curMask);	// Unmask..
			*(short *)(((char *)theSprite->maskPtr)+30) |= theSprite->curMask;		// Mask..
			}
		}

	// Now, move sprite Left or Right
	if ( theSprite->xOff != 0 )
		{
		short theMask = theSprite->curMask;
		
		// IF we're about to move to another mask space..
		if (((short *)theSprite->Pic)[0] == 0 ||
			((short *)theSprite->Pic)[1] == 0 )
			{
			// Test for a bounce of another mask...
			if ( (theMask == 0x6000) && (theSprite->xOff < 0) )	// Reached the side
				theSprite->xOff = 1;
			else if ( (theMask == 0x0003) && (theSprite->xOff > 0) )
				theSprite->xOff = -1;
			else
				{
				// Shift mask to new position
				theMask = (theSprite->xOff < 0) ? (theMask << 1) : (theMask >> 1);
				// Strip off overlapping bit
				theMask &= ~(theSprite->curMask);
				
				walk = theSprite->maskPtr;
				for ( idx = 0; idx < 16; idx++, walk++ )	// Run down the length of the icon testing for a used mask
					if ( (*walk & theMask) != 0 )			// Something already there, bounce!
						{
						theSprite->xOff = -(theSprite->xOff);
						break;
						}
				
				if ( idx >= 16 )
					{
					// Shift mask properly
					walk = theSprite->maskPtr;
					theMask = (theSprite->xOff < 0) ? (theSprite->curMask << 1) : (theSprite->curMask >> 1);
					theMask ^= theSprite->curMask;
					
					for ( idx = 0; idx < 16; idx++, walk++ )
						*walk ^= theMask;							// Remove old mask & set new one
					
					theSprite->curMask ^= theMask;
					}
				}
			
			}
		
		// Mask is now correct, shift screen and icon..
		
		longwalk = theSprite->Pic;
		if ( theSprite->xOff < 0 )
			{
			if ( ((short *)longwalk)[1] == 0 )
				{
				((char *)theSprite->ScreenPtr) -= 2;
				((char *)theSprite->BackPtr)   -= 2;
				for (idx=0;idx<17;idx++, longwalk++)
					*longwalk = *(unsigned short *)longwalk;		// Shift upper word into lower word
				}
			
			longwalk = theSprite->Pic;
			for (idx=0;idx<17;idx++, longwalk++)
				*longwalk <<= 1;
			}
		else
			{
			if ( ((short *)longwalk)[0] == 0 )
				{
				((char *)theSprite->ScreenPtr) += 2;
				((char *)theSprite->BackPtr)   += 2;
				for (idx=0;idx<17;idx++, longwalk++)
					*longwalk <<= 16;		// Shift lower word into upper word
				}
			
			longwalk = theSprite->Pic;	
			if ( ((short *)longwalk)[1] == 0 )
				{
				// Shift with mask for the first round...
				for (idx=0;idx<17;idx++, longwalk++)
					*longwalk = 0x7FFFFFFF & (*longwalk >> 1);
				}
			else
				{
				for (idx=0;idx<17;idx++, longwalk++)
					*longwalk >>= 1;
				}
			}
		}
	
	
}

void DrawSprite( Sprite *theSprite )
{
	long idx, *walk, *back, *sprite;
	walk = theSprite->ScreenPtr;
	back = theSprite->BackPtr;
	sprite = theSprite->Pic;
	sprite++;
	for (idx=0;idx<16;idx++, (char *)walk += 30, (char *)back += 30)
		*walk = (*back) | (*(sprite++));
}

void SpriteProc(long numArgs, long initLine)
{
	Sprite theSprite;
	long i, *walk;
	ImportantStuff *theStuff;

	if (numArgs != 1)
		return;

	theStuff = (ImportantStuff *)Xinu_Receive(0);
	
	// Initialize the Sprite Struct : recieve the screen buffer from calling proc
	{ // BT: Silly way to avoid multiplication
	short t=initLine;
	i=0;
	while ( t-- )
		i+=30;
	}
	theSprite.ScreenPtr = (long *)((char *)(theStuff->screenBuff) + i );
	theSprite.BackPtr = (long *)(0x4440 + i);	// --- NOT ROM 2.1 COMPATABLE
	theSprite.xOff = 1;
	theSprite.yOff = 1;
	theSprite.sleepTime = (initLine == 0) ? 0 : 1;
	theSprite.maskPtr = ((short *)theStuff->bufferMask) + initLine;
	theSprite.curMask = 0x6000;
	
	// --- BT: allow there to be conflicts, and resolve them!
	// Setup the initial mask (and hope there's nothing in the way!)
	for ( i = 0; i < 16; i++ )
		theSprite.maskPtr[i] |= theSprite.curMask;
		
	// Copy Sprite data into Sprite Struct
	walk =	(initLine == 0) ?
			((long *)XinuSprite) :
			((long *)TISprite);
	
	for ( i = 0; i < 17; i++ )
		theSprite.Pic[i] = *(walk++);
		
	while (1)
		{
		// Display sprite
		DrawSprite( &theSprite );

		// Give CPU to another process
		Xinu_Sleep(theSprite.sleepTime);
		
		Xinu_Wait( theStuff->maskSem );	// Enter mutually-excluded section of code
		MoveSprite( &theSprite );		// Move the sprite to it's new position...
		Xinu_Signal( theStuff->maskSem );	// Release mutual-exclusion
		}
}

void DrawWindowFrame( XinuWindow *theWind, char active )
{
	short	*walk = theWind->screenPtr;
	short i,j;
	
	if ( active )
		{
		for ( i=0; i<4; i++, (char *)walk += 30 )
			for ( j=0; j<theWind->width; j++ )
				walk[j] = 0xFFFF;					// Hilight 'title' bar
		}
	else
		{
		for ( i=0; i<4; i++, (char *)walk+=30 )
			{
			for ( j=0; j<theWind->width; j++ )		// Regular 'title' border
				{
				switch (i)
				{
				case 0:
				case 3:
					walk[j] = 0xFFFF;
					break;
				default:
					walk[j] = 0;
				}
				}
			
			*(char *)walk |= 0x80;
			walk[theWind->width-1] |= 1;
			}
		}
	
	for ( i=0; i<(theWind->height-4); i++, (char *)walk+=30 )
		{
		((char *)walk)[0] |= 0x80;
		walk[theWind->width-1] |= 1;
		}
	
	for ( j=0; j<theWind->width; j++ )
		*(walk++) = 0xFFFF;
}

void MoveWindow( XinuWindow *theWind, short dir )
{
	// First, move window up or down...
	if ( dir & 10 )	// If bits for UP or DOWN are set..
		{
		// Test for another object...
		if ( dir & 2 )	// Up..
			{
			if ( ((*(short *)(((char *)theWind->maskPtr)-2)) & (theWind->windowMask)) != 0 )
				dir &= 0xFFFD;	// Turn off UP bit..
			}
		else if ( ((*(short *)(theWind->maskPtr + theWind->height)) & (theWind->windowMask)) != 0 )
			dir &= 0xFFFF7;		// Turh off DOWN bit..
		
		
		if ( dir & 2 )	// Up..
			{
			(char *)theWind->screenPtr	-= 30;					// Adjust drawing pointer
			*(short *)(((char *)theWind->maskPtr)-2)  |= theWind->windowMask;
			*(short *)(theWind->maskPtr+theWind->height) &= ~(theWind->windowMask);
			(char *)theWind->maskPtr	-= 2;					// Adjust mask pointer
			theWind->y--;
			}
		else			// Down..
			{
			(char *)theWind->screenPtr	+= 30;					// Adjust drawing pointer
			(char *)theWind->maskPtr	+= 2;					// Adjust mask pointer
			*(short *)(((char *)theWind->maskPtr)-2)  &= ~(theWind->windowMask);	// Unmask..
			*(short *)(theWind->maskPtr+theWind->height) |= theWind->windowMask;		// Mask..
			theWind->y++;
			}
		}
	
	//
	// Move window Right/Left 16 pixels at a time...
	//
	
	// TO DO
}

void DrawContents( ImportantStuff *theStuff, XinuWindow *theWind, char active, char *theStr )
{
	short ch,i,*walk,*walk2;
	
	// Store TI-LCD contents, clear space on main LCD
	walk = (short *)((char *)theWind->screenPtr - (char *)theStuff->screenBuff + (char *)0x4440);	// --- NOT ROM 2.1 Compatable
	walk2 = theWind->screenPtr;
	for ( i = 0; i < theWind->height; i++, (char *)walk += 30, (char *)walk2 += 30 )
		for ( ch = 0; ch < theWind->width; ch++ )
			{
			walk2[ch] = walk[ch];
			walk[ch] = 0;
			}
	
	// Draw string onto workspace in main LCD
	TI_FontSetSys( MEDIUM );
	TI_DrawStrXY( theWind->x, theWind->y, theStr, active ? PLAIN : DIMMED );
	
	// Swap Contents of workspace, replacing TI-LCD contents
	walk = (short *)((char *)theWind->screenPtr - (char *)theStuff->screenBuff + (char *)0x4440);	// --- NOT ROM 2.1 Compatable
	walk2 = theWind->screenPtr;
	for ( i = 0; i < theWind->height; i++, (char *)walk += 30, (char *)walk2 += 30 )
		for ( ch = 0; ch < theWind->width; ch++ )
			{
			short x;
			x = walk[ch];
			walk[ch] = walk2[ch];
			walk2[ch] = x;
			}

}

void WindowProc()
{
	XinuWindow theWind;
	ImportantStuff *theStuff;
	short i,ch;
	char active,theStr[6]="Xinu!";
	
	theStuff = (ImportantStuff *)Xinu_Receive(0);

	theWind.height = 16;
	theWind.width = 3;
	theWind.x = 24;
	theWind.y = 66;
	theWind.windowMask = 0x3800;
	theWind.screenPtr = (short *)((char *)(theStuff->screenBuff) + 30*60 + 2 );
	theWind.maskPtr = ((short *)theStuff->bufferMask) + 60;
	
	
	// --- TO DO: allow there to be conflicts, and resolve them
	// Setup the initial mask (and hope there's nothing in the way!)
	for ( i = 0; i < theWind.height; i++ )
		theWind.maskPtr[i] |= theWind.windowMask;
	
	active = 0;
	DrawContents( theStuff, &theWind, active, theStr );
	DrawWindowFrame( &theWind, active );
	while (1)
		{
		if ( (ch = Xinu_GetKey()) > 0 )			// Get a keystroke from the system..
			{
			if (!active)
				{
				active = 1;
				
				// Just in case it was lost, reset LCD buffer to Xinu Overlay Buffer
				*((short *)0x600010) = ((long)theStuff->screenBuff) >> 3;
				}
			
			switch (ch)
			{
			case 0x0151:	// Left
			
				Xinu_Wait( theStuff->maskSem );	// Enter mutually-excluded section of code
				MoveWindow( &theWind, 1 );		// Move the window to it's new position...
				Xinu_Signal( theStuff->maskSem );
				break;
			
			case 0x0152:	// Up
			
				Xinu_Wait( theStuff->maskSem );	// Enter mutually-excluded section of code
				MoveWindow( &theWind, 2 );		// Move the window to it's new position...
				Xinu_Signal( theStuff->maskSem );
				break;
			
			case 0x0154:	// Right
			
				Xinu_Wait( theStuff->maskSem );	// Enter mutually-excluded section of code
				MoveWindow( &theWind, 4 );		// Move the window to it's new position...
				Xinu_Signal( theStuff->maskSem );
				break;
			
			case 0x0158:	// Down
			
				Xinu_Wait( theStuff->maskSem );	// Enter mutually-excluded section of code
				MoveWindow( &theWind, 8 );		// Move the window to it's new position...
				Xinu_Signal( theStuff->maskSem );
				break;
			
			default:		// Any other character...
				for (i=0;i<5;i++)
					theStr[i]=theStr[i+1];			// Scroll String array
	
				theStr[4] = ch;						// Insert new char
				theStr[5] = 0;
			}
			
			DrawContents( theStuff, &theWind, active, theStr );
			DrawWindowFrame( &theWind, active );
			}
		else if ( ch < 0 && active )
			{
			active = 0;
			DrawContents( theStuff, &theWind, active, theStr );
			DrawWindowFrame( &theWind, active );
			}
		
		Xinu_Sleep(0);
		}
	
}

void FinalDemo()
{
	ImportantStuff theStuff;
	PROC_ID theProc;
	
	theStuff.screenBuff = (short *)((((long)ScreenBuffer)+8) & 0xFFFFFFF8);
	theStuff.bufferMask = (short *)((char *)BufferMask + 2);
	theStuff.maskSem    = Xinu_SemCreate( 1 );
	
	{	// Clear screen mask area...
	register temp;
	for ( temp=0; temp<64; temp++ )
		((long *)theStuff.bufferMask)[temp] = 0;
	}
	
	// Set the screen to the new screen buffer...
	*((short *)0x600010) = ((long)theStuff.screenBuff) >> 3;
	
	// Create Sprite Procs, send them the important stuff
	Xinu_Resume( theProc = Xinu_CreateProc((void *)SpriteProc,800,10,1,0) );
	Xinu_Send( theProc, (WORD)&theStuff );
	
	Xinu_Resume( theProc = Xinu_CreateProc((void *)SpriteProc,800,10,1,30) );
	Xinu_Send( theProc, (WORD)&theStuff );
	
	// Create Window Procs, send them the important stuff too...
	Xinu_Resume( theProc = Xinu_CreateProc((void *)WindowProc,800,10,1,0) );
	Xinu_Send( theProc, (WORD)&theStuff );
	
	while (1)
		{
		CopyScreen( (char *)theStuff.bufferMask, (char *)theStuff.screenBuff );	// Do nothing forever!
		Xinu_Sleep(0);
		}
}

typedef struct
	{
	char name[8];
	short flags;
	short handle;
	} SymEntry;

//void doDemo()
void myCode()
{
	long data;
	
	/*
	LoadStruct *theStuff;
	SymEntry *theSym;
	PROC_ID theProc;
	short folderSym;
	char *begin;
	
	
	theStuff = (LoadStruct *)Xinu_GetMem( 8 );
	
	folderSym = *((short *)(TI_main_lcd+0x194C));
	Xinu_kprintf( "Hand=%lx", folderSym );
	
	theSym = (SymEntry *)FindSymEntry( folderSym, "main" );
	
	Xinu_kprintf( "Hand=%lx", theSym->handle );
	
	theSym = (SymEntry *)FindSymEntry( theSym->handle, "xinudemo" );
	
	theStuff->theHandle = theSym->handle;
	
	begin = (char *)((*(long **)(TI_main_lcd + 0x1902))[theStuff->theHandle]);
	theStuff->beginOffset = (short)((char *)FinalDemo - begin);
	
	Xinu_kprintf( "Hand=%lx, Begin=%lx  ", theSym->handle, begin );
	
	theProc = Xinu_LoadProc();
	Xinu_Send( theProc, (WORD)theStuff );
	
// ------------ For Debugging Purposes Only ---------
	//enable();							// Allow 
	*((short *)(0x75AA+6)) = 0;			// --- NOT ROM 2.1 compatable
	while ( *((short *)(0x75AA+6)) == 0 )
		; 								// Do nothing until keypress
	*((short *)(0x75AA+6)) = 0;
// --------------------------------------------------
	*/
	
	// BT: Load the library, and it's first function will be called..
	//Xinu_kprintf( "Loading...%lx", (long)myCode );
	if ( Xinu_LoadLibrary( "xinudemo", &data ) == 0 )
		{
		//Xinu_kprintf( "Running..." );
		FinalDemo();	// IF we're already loaded, run the demo!
		//while (1)
		//	*(char *)TI_main_lcd = ~(*(char *)TI_main_lcd);
		}
	
	//Xinu_kprintf( "Quitting...%lx", (long)myCode );
}

//*/
