
#include <Globals.h>
#include <TIOS.h>

// ----------------------------------------------------------------
// BT: Testing: Fargo interface to start a process without crashing
// ----------------------------------------------------------------

typedef struct	// Structure to pass which file to load
	{
	short	theHandle;		// Handle of the data to copy
	short	beginOffset;	// Offset of process to run
	char	**theLibData;	// Place to store the loaded address, 0 == ignore
	} LoadStruct;

typedef struct	// Structure which replaces the loaded file's data
	{
	short   theLen;			// Length Field (TI-OS)
	short	theHeader;		// Header == 0x0032 if FARGO II file
	long	theType;		// 'EXE ' for FARGO app, 'XINU' for XINU Library
	long	theType2;		// 'APPL' for FARGO app, 'isUP' for loaded XINU Library
	char	*theLibrary;	// Pointer to the library in Xinu space
	} LoadedStruct;

typedef struct
	{
	char name[8];
	short flags;
	short handle;
	} SymEntry;
	
void LoaderProc()
{
	LoadStruct *theStuff;
	long *begin, *walk, *theMem, size;
	
	LoadGlobals();
	
	while (1)
		{
		theStuff = (LoadStruct *)receive(0);
		
		begin = (long *)((*(long **)(TI_main_lcd + 0x1902))[theStuff->theHandle]);
		size = *(short *)begin + 4;
		
		//kprintf( "Begin=%lx,Sz=%lx", begin, size );
		
		// Allocate Xinu Non-Relocatable Memory
		theMem = (long *)getmem( size );
		
		//kprintf( "theMem=%lx", (long)theMem );
		// Copy demo to allocated memory
		size >>= 2;		// Convert to # of longwords
		walk = theMem;
		while ( size-- )
			*(walk++) = *(begin++);
			
		if ( theStuff->theLibData )	// Store the loaded address, if requested
			*(theStuff->theLibData) = (char *)theMem;
		
		// Delete the old handle, replacing it with a small struct..
		
		//if ( TI_HeapRealloc(theStuff->theHandle, sizeof(LoadedStruct)) != 0 )
			{
			begin = (long *)((*(long **)(TI_main_lcd + 0x1902))[theStuff->theHandle]);
			((LoadedStruct *)begin)->theHeader = 0x0032;
			((LoadedStruct *)begin)->theType = 'XINU';
			((LoadedStruct *)begin)->theType2 = 'isUP';
			((LoadedStruct *)begin)->theLibrary = (char *)theMem;
			//kprintf( "OK!" );
			}
		
		//else
		//	kprintf( "BAD!" );	// An error occured during relocation.. ?
		
		(char *)theMem += theStuff->beginOffset;
		freemem( (struct mblock *)theStuff, sizeof(LoadStruct) );	// Release the used memory
		
		//kprintf( "BeginOffset=%lx", theStuff->beginOffset );
		//kprintf( "BeginData=%lx", *theMem );
		//kprintf( "About to Create Proc!" );
		
		// Now create the process that will run the loaded program..
		ready( create((void *)theMem,800,10,1,0), 0 );
		}
}

short LoadLibrary( char *LibName, char **LibData )
{
	//
	// Loads library named 'LibName' into Xinu Memory and stores references to it at LibData
	//
	
	// Find library in TI-Memory
	//    Copy library to Xinu Memory
	// If not there, find library in Xinu Memory
	// Put pointer to library jump table in LibData field
	
	short folderSym;
	SymEntry *theSym;
	LoadStruct *theStuff;
	LoadedStruct *isUp;
	char *begin;
	short sr, ret = 0; 
	long a4;
	
	a4 = SaveA4AndLoadGlobals();
	sr = disable();
	
	theStuff = (LoadStruct *)getmem( sizeof(LoadStruct) );
	//kprintf( "Loading Library,sadr=%lx", (long)theStuff );
	folderSym = *((short *)(TI_main_lcd+0x194C));
	theSym = (SymEntry *)TI_FindSymEntry( folderSym, "main" );
	//kprintf( "Main Handle: %lx", theSym->handle );
	theSym = (SymEntry *)TI_FindSymEntry( theSym->handle, LibName );
	//kprintf( "Lib Handle: %lx", theSym->handle );
	theStuff->theHandle = theSym->handle;
	begin = (char *)((*(long **)(TI_main_lcd + 0x1902))[theStuff->theHandle]);
	
	//kprintf( "Lib Ptr: %lx", begin );
	isUp = (LoadedStruct *)begin;
	
	if ( (isUp->theType == 'XINU') && (isUp->theType2 == 'isUP') )
		{
		//kprintf( "Already Loaded!  " );
		*LibData = isUp->theLibrary;		// Already loaded, just copy data.
		}
	else
		{
		//kprintf( "First Load! %s, %s ", (char *)&isUp->theType, (char *)&isUp->theType );
		theStuff->beginOffset = ((short *)(begin + *(short *)(begin+0x14)))[1];
		//theStuff->beginOffset = ((short *)(begin+0x14))[1];	// First exported symbol's offset (from FARGO II format object file)
		theStuff->theLibData = LibData;		// Pass on mark location to loader proc
		*(LibData) = 0;						// Set to zero to show it's not valid yet!
		
		send( xinuGlobs.LOADERPROC, (WORD)theStuff );
		ret = 1;							// Return 1 if loading...
		}
	
	RestoreA4AndSR(a4,sr);
	return ret;								// Return 0 if loaded..
}

short UnloadLibrary( long *LibData )
{
	//
	// Unloads Library
	//
	return -1;
}