/* initialize.c - nulluser, sysinit */

#include <Globals.h>
#include <TIOS.h>

extern	int	start();			/* startup code */
extern	void Trap14Exception();	// System Call Trap Processor
extern	void nullproc();
extern	void LoaderProc();

extern	void FinalDemo();		// The demo program

/************************************************************************/
/***				NOTE:				      						  ***/
/***								     							  ***/
/***   This is where the system begins after the C environment has    ***/
/***   been established.  Interrupts are initially DISABLED, and      ***/
/***   must eventually be enabled explicitly.  This routine turns     ***/
/***   itself into the null process after initialization.  Because    ***/
/***   the null process must always remain ready to run, it cannot    ***/
/***   execute code that might cause it to be suspended, wait for a   ***/
/***   semaphore, or put to sleep, or exit.  In particular, it must   ***/
/***   not do I/O unless it uses kprintf for polled output.           ***/
/***								      							  ***/
/************************************************************************/

/*------------------------------------------------------------------------
 *  sysinit  --  initialize all Xinu data structeres and devices
 *------------------------------------------------------------------------
 */

void sysinit()
{

	PROC_ID	tempProc;
	short	i;
	Process	*pptr;
	QItem	*sptr;
	
	//kprintf(" Xinu %s ", VERSION);			// --- DEBUGGING

	initevec();								// Install trap handlers
	nextproc = NPROC-1;
	nextqueue = 0;							// New Queue system
	
	// 	Initialize free memory list
	// --- Insert code here to initialize Xinu Dynamic Memory
	
	pptr = xinuGlobs.theProcs;
	for (i=0; i<NPROC; i++, pptr++)			// initialize process table
		pptr->pstate = PRFREE;

	if ((tempProc = create((void *)nullproc,400,10,1,0)) == (PROC_ID)SYSERR)
		kprintf("Error creating NULLPROC!");// BT: Create the null process
	pptr = (Process *)tempProc;				// BT: Edit the null process		
	pptr->pprio = 0;						// Has zero priority
	xinuGlobs.NULLPROC = tempProc;			// BT: Save null proc's pid
	
	// BT:
	// The Fargo program currently running will become the TI-OS process.
	// First, we create a Xinu process descriptor.
	// Then delete the stack (TI-OS has one already).
	// After restoring interrupts, we will leave the Xinu startup to return
	//    to the TI-OS.
	// Eventually, the preemption interrupt will occur, and the current
	//    state will be saved as well as the stack (it will be the user
	//    stack hopefully).  Xinu will select another process to run,
	//    probably the startup proc, if any, or TI-OS again.
	//

	if ((tempProc = create((void *)nullproc,0,10,1,0))==(PROC_ID)SYSERR)
		kprintf("Error creating TIPROC!");	// BT: Create the TI-OS process	
	
	pptr = (Process *)tempProc;
	xinuGlobs.TIPROC = tempProc;			// BT: Save TI Proc's pid
	freestk(pptr->pbase,pptr->pstklen);		// BT: Free the stack (TI-OS has it's own stack)
	pptr->pstate = PRCURR;					// It's the current one!
	pptr->pbase = 0x4440;					// Setup actual TI stack info --- NOT ROM 2.1 compatable
	//pptr->plimit = 0x120;
	pptr->pstklen = 0x4440-0x120;			// --- NOT ROM 2.1 compatable
	xinuGlobs.CurPID = tempProc;			// Current process is this one!

	sptr = xinuGlobs.theQs;
	for (i=0; i<NQENT; i++, sptr++)		// Initialize Queues
		{
		sptr->next = sptr->prev = 0;
		sptr->flags = Q_FLAG;
		}

	xinuGlobs.readyQ = NewQueue();		// initialize ready list
	ready( xinuGlobs.NULLPROC, 0 );		// Ready nullproc to run

	set_evec( 0xB8, Trap14Exception );	// BT: Install the trap processor for system calls
	clkinit();							// initialize real time clock
	
	ready( tempProc = create((void *)LoaderProc,800,10,1,0), 0 );
	xinuGlobs.LOADERPROC = tempProc;

	// BT: we now 'return', and become the TI-OS process!
	return;
}
