/* resched.c  -  resched */

#include <Globals.h>


/*------------------------------------------------------------------------
 * resched  --  reschedule processor to highest priority ready process
 *
 * Notes:	Upon entry, currpid gives current process id.
 *		Proctab[currpid].pstate gives correct NEXT state for
 *			current process if other than PRREADY.
 *------------------------------------------------------------------------
 */


asm long getSP()	// BT
{
	move.l	sp,d0
	rts
}
/*
asm short getSR()	// BT --- DEBUG CODE
{
	move	sr,d0
	rts
}
*/

short resched( )
{
	Process *oldProc=(Process *)xinuGlobs.CurPID,*newProc;
	unsigned long currSP;			// REAL sp of current process - BT: was a global!!
	
	/* no switch needed if current process priority higher than next*/
	if ( (oldProc->pstate == PRCURR) &&
	   (lastkey(xinuGlobs.readyQ) < oldProc->pprio) )
		return 0;

#ifdef STKCHK
	// make sure CURRENT process is within it's stack
	
	currSP = getSP();	// BT: inline assembly not allowed.
	if ( (currSP < ((unsigned)oldProc->pbase - oldProc->pstklen)) ||
		 (currSP > (unsigned)oldProc->pbase) )	// BT: check for stack 'underflow' also...
		{
		//short sr = getSR();
		//kprintf( "Proc=%lx,Stk=%lx,Lim=%lx,SR=%lx ", (long)oldProc, currSP, oldProc->plimit, sr );
		panic(" CurProc stack over/under flow! ");
		}

#endif

	/* force context switch */

	if (oldProc->pstate == PRCURR)
		{
		oldProc->pstate = PRREADY;
		InsertByKey( (PROC_ID)oldProc, xinuGlobs.readyQ, oldProc->pprio);
		}

	/* remove highest priority process at end of ready list */
	newProc = (Process *)GetLast(xinuGlobs.readyQ);
	xinuGlobs.CurPID = (PROC_ID)newProc;
	newProc->pstate = PRCURR;		/* mark it currently running	*/
	
#ifdef	STKCHK
	if (( *((WORD *)newProc->pbase) != MAGIC ) &&
		((char *)newProc != (char *)xinuGlobs.TIPROC) )
			{
			panic( " Stack corrupted!" );
			//kprintf( " oP=%lx,oS=%lx,oL=%lx ", (long)oldProc, (long)(oldProc->pbase), (long)(oldProc->plimit)  );
			//kprintf( " nP=%lx, nS=%lx ", (long)newProc, (long)(newProc->pbase) );
			//ret_mon();
			}
	if ( ((unsigned)newProc->pcurstack) < ((unsigned)newProc->pbase - newProc->pstklen) ||
		 ((unsigned)newProc->pcurstack) > ((unsigned)newProc->pbase) ) // BT: check for stack 'underflow' also...
		{
		//kprintf( "CurProc=%lx, Stk=%lx, Lim=%lx", (long)newProc, newProc->pcurstack, newProc->plimit );
		panic(" Stack over/under flow! ");
		}
#endif
	
	xinuGlobs.preempt = QUANTUM;
	
	// BT: If the process was just freed, only load the new context, don't switch!
	if (oldProc->pstate == PRFREE)
		{
		ctxld( &(newProc->pcurstack) );							// Simply load the new proc & run!
		}
	else
		ctxsw( &(oldProc->pcurstack), &(newProc->pcurstack) );	// BT: Old Proc ends, New Proc begins...

	/* The OLD process returns here when resumed. */
	return 0;
}
